import platform
import subprocess
from ..constants import *
from ..layers import *
from ..utils import *
from ..materials import *
import json
import gdsfactory as gf
from copy import deepcopy

# from time import time
import datetime
from math import cos, pi, sin
import os
import numpy as np

from sortedcontainers import SortedDict, SortedSet
from gdsfactory.generic_tech import LAYER_STACK, LAYER


def setup(
    path,
    study,
    characteristic_wavelength,
    wavelengths,
    wl_1_f=None,
    bbox=None,
    boundaries=["PML", "PML", "PML"],
    nres=4,
    dx=None,
    dy=None,
    dz=None,
    component=None,
    z=None,
    # zmargin=None,
    zmin=None,
    zmax=None,
    zmargin_mode=None,
    xmargin_mode=None,
    inset=0,
    port_margin="auto",
    runs=[],
    sources=[],
    layer_stack=SOI220,
    materials=dict(),
    core="core",
    exclude_layers=[],
    Courant=None,
    gpu=None,
    dtype=np.float32,
    saveat=10,
    magic="",
    name=None,
    source_port_margin=0.1,
    modes=None,
    ports=None,
    approx_2D_mode=False,
    show_field="Hz",
    T=None,
    field_decay_threshold=None,
    path_length_multiple=None,
    relative_pml_depths=1,
    relcourant=0.9,
    hasPEC=None,
    keys=None,
    ordering="frequency",
    #
    field_slices=None,
    geometry_slices=None,
    force=False,
    verbose=False,
    pixel_size=0.01,
    secret=None,
    info=None,
):
    # if force:
    #     shutil.rmtree(path, ignore_errors=True)
    # elif os.path.exists(path):
    #     raise FileExistsError(
    #         f"Path {path} already exists. Use force=True to overwrite."
    #     )

    os.makedirs(path, exist_ok=True)
    if info is None:
        info = {}
    json.dump(info, open(os.path.join(path, "info.json"), "w"), indent=4)
    if approx_2D_mode:
        N = 2
    else:
        N = 3
        approx_2D_mode = None

    if inset is None:
        inset = [0] * N

    prob = {
        "nres": nres,
        "characteristic_wavelength": characteristic_wavelength,
        "wavelengths": wavelengths,
        "name": name,
        "path": path,
        "keys": keys,
        "show_field": show_field,
        "dx": dx,
        "dy": dy,
        "dz": dz,
        "z": z,
        "T": T,
        "field_decay_threshold": field_decay_threshold,
        "path_length_multiple": path_length_multiple,
        "saveat": saveat,
        "verbose": verbose,
        "field_slices": field_slices,
        "geometry_slices": geometry_slices,
        "boundaries": boundaries[0:N],
        "approx_2D_mode": approx_2D_mode,
        "relative_pml_depths": relative_pml_depths,
        "relcourant": relcourant,
        "hasPEC": hasPEC,
        "ordering": ordering,
        "pixel_size": pixel_size,
        "secret": secret,
    }

    prob["class"] = "pic"
    prob["dtype"] = str(dtype)
    prob["timestamp"] = (
        datetime.datetime.now().isoformat(timespec="seconds").replace(":", "-")
    )
    prob["magic"] = magic

    gpu_backend = gpu
    # if gpu_backend:s
    prob["gpu_backend"] = gpu_backend

    if component is None:
        0
    else:
        c = component
        if ports is None:
            ports = []

        if bbox is None:
            bbox = c.bbox_np()
        l, w = bbox[1] - bbox[0]
        for v in ports:
            if "layer" in v:
                0
            else:
                if "z" in v:
                    v["center"] = (sum(bbox) / 2).tolist() + [v["z"]]
                x, y = v["center"][:2]
                if "length" not in v:
                    v["length"] = l
                if "width" not in v:
                    v["width"] = w
                v["start"] = [-v["length"] / 2, -v["width"] / 2]
                v["stop"] = [+v["length"] / 2, +v["width"] / 2]
                if v["direction"] == "+z":
                    frame = [[1, 0, 0], [0, 1, 0], [0, 0, 1]]
                elif v["direction"] == "-z":
                    frame = [[-1, 0, 0], [0, 1, 0], [0, 0, -1]]
                v["frame"] = frame
                v["normal"] = frame[2]
                v["tangent"] = frame[0]

        for p in c.get_ports_list(prefix="o"):
            v = {
                "name": p.name,
                "center": (np.array(p.center) / 1e0).tolist(),
                "normal": [
                    cos(p.orientation / 180 * pi),
                    sin(p.orientation / 180 * pi),
                ],
                "tangent": [
                    -sin(p.orientation / 180 * pi),
                    cos(p.orientation / 180 * pi),
                ],
                "width": p.width / 1e0,
            }
            z, n = [0, 0, 1], [*v["normal"], 0]
            t = np.cross(z, n).tolist()
            v["frame"] = [t, z, n]
            ports.append(v)

        prob["ports"] = ports

        bbox = bbox.tolist()
        bbox[0].append(zmin)
        bbox[1].append(zmax)
        print(f"bbox: {bbox}")
        prob["bbox"] = bbox

        if c and c.get_ports_list():
            d = layer_stack.layers[core]
            hcore = d.thickness
            zcore = d.zmin

            if zmargin_mode is None:
                zmargin_mode = 3 * hcore
            if type(zmargin_mode) in [int, float]:
                zmargin_mode = [zmargin_mode, zmargin_mode]
            # if zmargin is None:
            #     zmargin = [zmargin_mode[0], zmargin_mode[1]]
            # if type(zmargin) in [int, float]:
            #     zmargin = [zmargin, zmargin]

            port_width = max([p.width / 1e0 for p in c.ports])
            if xmargin_mode is None:
                xmargin_mode = port_width
            if type(xmargin_mode) in [int, float]:
                xmargin_mode = [xmargin_mode, xmargin_mode]
            print(f"xmargin_mode: {xmargin_mode}")

            if zmin is None:
                zmin = zcore - 3 * hcore
            if zmax is None:
                zmax = zcore + 4 * hcore

            # h = hcore + zmargin[0] + zmargin[1]
            # zmin = zcore - zmargin[0]
            # zmax = zmin + h
            #
            # xmargin = ymargin = 2*port_width

            wmode = port_width + xmargin_mode[0] + xmargin_mode[1]
            hmode = hcore + zmargin_mode[0] + zmargin_mode[1]
            zmode = zcore - zmargin_mode[0]
            zcenter = zmode + hmode / 2

        layers = set(c.layers) - set(exclude_layers)

        if modes is None:
            modes = [
                {
                    "wavelength_range": [wavelengths[0], wavelengths[-1]],
                    "nmodes": 1,
                }
            ]
        for mode in modes:
            # if "wavelength" in mode:
            #     assert "frequency" not in mode
            #     mode["frequency"] = wl_1_f / mode["wavelength"]
            #     if "bandwidth" in mode:
            #         mode["bandwidth"] *= mode["frequency"] / mode["wavelength"]
            # if "frequency" in mode:
            #     assert "wavelength" not in mode
            #     mode["wavelength"] = wl_1_f / mode["frequency"]
            #     if "bandwidth" in mode:
            #         mode["bandwidth"] *= mode["wavelength"] / mode["frequency"]
            if "frequency_range" in mode:
                mode["wavelength_range"] = [
                    wl_1_f / mode["frequency_range"][1],
                    wl_1_f / mode["frequency_range"][0],
                ]
            if "wavelength_range" not in mode:
                mode["wavelength_range"] = [wavelengths[0], wavelengths[-1]]
            if "ports" not in mode:
                mode["ports"] = None

        MODES = os.path.join(path, "modes")
        os.makedirs(MODES, exist_ok=True)
        GEOMETRY = os.path.join(path, "geometry")
        os.makedirs(GEOMETRY, exist_ok=True)

        layer_stack_info = material_voxelate(
            c, zmin, zmax, layers, layer_stack, materials, GEOMETRY
        )

        dir = os.path.dirname(os.path.realpath(__file__))

        for f in ["solvemodes.py"]:
            fn = os.path.join(dir, f)
            if platform.system() == "Windows":
                os.system(f"copy /Y {fn} {MODES}")
            else:
                subprocess.run(["cp", fn, MODES])
        prob["layer_stack"] = layer_stack_info
        prob["study"] = study
        prob["materials"] = materials

        prob["N"] = N

        for run in runs:
            for k, v in run["sources"].items():
                # p = ports[k]
                p = next((p for p in ports if p["name"] == k), None)
                ct = np.array(p["center"])
                n = np.array(p["normal"])
                v["center"] = (ct + n * source_port_margin).tolist()

                # modes = solve_modes(polys, eps, bbox)
            for k, v in run["monitors"].items():
                p = next((p for p in ports if p["name"] == k), None)
                v["center"] = copy.deepcopy(p["center"])
            for k, v in list(run["sources"].items()) + list(run["monitors"].items()):
                p = next((p for p in ports if p["name"] == k), None)
                v["frame"] = p["frame"]
                if len(v["center"]) == 2:
                    v["center"] += [zcenter]
                    v["start"] = [-wmode / 2, zmode - zcenter]
                    v["stop"] = [wmode / 2, zmode + hmode - zcenter]
                else:
                    v["start"] = p["start"]
                    v["stop"] = p["stop"]
        bg = materials["background"]["epsilon"]
        ime = []
        hasPEC = False
        for f in os.listdir(GEOMETRY):
            i, mat, _ = f[:-4].split("_")
            if mat != "design" and mat != "None":
                if mat != "PEC":
                    eps = materials[mat]["epsilon"]
                else:
                    hasPEC = True
                    eps = "PEC"
                ime.append((int(i), trimesh.load(os.path.join(GEOMETRY, f)), eps))
        mesheps = [x[1:] for x in sorted(ime, key=lambda x: x[0])]
        l = []
        for mode in modes:
            if not mode["ports"]:
                v = runs[0]["monitors"].values()[0]
            else:
                v = runs[0]["monitors"][mode["ports"][0]]
            start = v["start"]
            stop = v["stop"]

            L = stop[0] - start[0]
            W = stop[1] - start[1]
            dl = mode.get("dx", L / 50)
            dx = L / round(L / dl)
            dy = W / round(W / dl)

            if "frequency" not in mode and "wavelength" not in mode:
                mode["wavelength"] = characteristic_wavelength
            elif "frequency" in mode:
                mode["wavelength"] = wl_1_f / mode["frequency"]

            if "modes" in mode:
                mode["dl"] = [dx, dy]
                for _mode in mode["modes"]:
                    for k, v in _mode.items():
                        if type(v) is np.ndarray:
                            _mode[k] = v.tolist()
                l.append(mode)
            else:
                polyeps = section_mesh(
                    mesheps,
                    v["center"],
                    v["frame"],
                    start,
                    stop,
                    bg,
                )
                # print(f"polyeps: {polyeps}")
                if "PEC_boundaries" in mode:
                    hasPEC = True
                    PEC_boundaries = mode["PEC_boundaries"]
                else:
                    PEC_boundaries = []
                nmodes = mode.get("nmodes", 1)

                wavelength = mode["wavelength"]
                eps = epsilon_from_polyeps(polyeps, bg, start, stop, dx, dy).tolist()
                if hasPEC:
                    _modes = solvemodes_femwell(
                        polyeps,
                        bg,
                        start,
                        stop,
                        wavelength,
                        nmodes,
                        dx,
                        dy,
                        PEC_boundaries,
                    )
                else:
                    _modes = solvemodes(
                        polyeps, bg, start, stop, wavelength, nmodes, dx, dy
                    )
                l.append(
                    {
                        "wavelength_range": mode["wavelength_range"],
                        "modes": _modes,
                        "ports": mode["ports"],
                        "dl": [dx, dy],
                        "epsilon": eps,
                    }
                )
        prob["modes"] = l
        prob["runs"] = runs

    if not os.path.exists(path):
        os.makedirs(path)
    return prob


def port_name(port):
    s = str(port).split("@")[0]
    if s[0] == "o":
        return s
    return f"o{s}"


def port_number(port):
    s = str(port).split("@")[0]
    if s[0] == "o":
        s = s[1:]
    return int(s)


def mode_number(port):
    l = str(port).split("@")
    return 0 if len(l) == 1 else int(l[1])


def unpack_sparam_key(k):
    o, i = k.split(",")
    po, pi = port_name(o), port_name(i)
    mo, mi = mode_number(o), mode_number(i)
    return po, mo, pi, mi


def long_sparam_key(k):
    po, mo, pi, mi = unpack_sparam_key(k)
    return f"{po}@{mo},{pi}@{mi}"



import numpy as np
import EMpy
import os
import sys
import json

path = sys.argv[1]
# data = np.load(os.path.join(path, "args.npz"))
data = json.loads(open(os.path.join(path, "args.json"), "rb").read())
eps = data["eps"]
eps = np.array(eps).T
dl = data["dl"]
λ = data["characteristic_wavelength"]
neigs = data["neigs"]
name = data["name"]
is2d = data["is2d"]

m, n = eps.shape
# print(m, n)

m += 1
n += 1
x = np.linspace(0.5 * dl, (m - 0.5) * dl, m)
y = np.linspace(0.5 * dl, (n - 0.5) * dl, n)


def ϵfunc(x_, y_):
    return eps


tol = 1e-6
if is2d:
    print("2d")
    solver = EMpy.modesolvers.FD.VFDModeSolver(λ, x, y, ϵfunc, "AA00").solve(
        2 * neigs, tol
    )
    modes = solver.modes
else:
    solver = EMpy.modesolvers.FD.VFDModeSolver(λ, x, y, ϵfunc, "0000").solve(neigs, tol)
    modes = solver.modes
    # modes = sorted(solver.modes, key=lambda x: -np.abs(x.neff))


neffs = [np.real(m.neff) for m in modes]
modes = [
    {k: m.get_field(k, x, y) for k in ["Ex", "Ey", "Ez", "Hx", "Hy", "Hz"]}
    for m in modes
]
for i, mode in enumerate(modes):
    np.savez(os.path.join(path, f"{name}_mode_{i}.npz"), **modes[i])



from ..materials import *
from .sparams import *
from .setup import *
from ..constants import *
from ..layers import *
from ..utils import *
import gdsfactory as gf
from copy import deepcopy
from functools import partial
from math import cos, pi, sin
import os
from gdsfactory.generic_tech import LAYER_STACK, LAYER
import json


def make_design_prob(
    path,
    component,
    targets,
    lmin,
    iters=10,
    # lvoid=None,
    # lsolid=None,
    symmetries=[],
    weights=dict(),
    eta=0.4,
    init=1,
    stoploss=None,
    layer=DESIGN,
    fill_material=None,
    void_material=None,
    materials={},
    #  fill_layer=LAYER.WG,
    #  void_layer=None,
    layer_stack=SOI220,
    restart=True,
    save_memory=False,
    **kwargs,
):
    assert (
        "field_decay_threshold" not in kwargs
    ), "field_decay_threshold is not supported in inverse design"
    materials = {**MATERIALS, **materials}
    c = component
    # if not N:
    #     raise NotImplementedError(
    #         "3D inverse design feature must be requested from Luminescent AI info@luminescentai.com")

    c.flatten()
    prob = make_prob(
        path,
        component=c,
        layer_stack=layer_stack,
        materials=materials,
        wavelengths=wavelengths,
        study="inverse_design",
        keys=keys,
        path_length_multiple=path_length_multiple,
        **kwargs,
    )

    prob["restart"] = restart
    prob["weights"] = {
        **{
            "tparams": 1,
            "sparams": 1,
            "phasediff": 1,
        },
        **weights,
    }
    prob["save_memory"] = save_memory

    # prob["lvoid"] = lvoid
    # prob["lsolid"] = lsolid
    prob["lmin"] = lmin
    prob["stoploss"] = stoploss
    prob["design_config"] = dict()
    # l = get_layers(layer_stack, fill_layer)[0]
    # d = {"thickness": l.thickness,
    #      "material": matname(l.material), "zmin": l.zmin}
    # d["layer"] = fill_layer
    # prob["design_config"]["fill"] = d

    # if void_layer is not None:
    #     l = get_layers(layer_stack, void_layer)[0]
    #     d = {"thickness": l.thickness, "material": l.material, "zmin": l.zmin}
    #     d["layer"] = void_layer
    #     d["epsilon"] = materials[d["material"]]["epsilon"]
    #     d["epsilon"] = d["epsilon"]
    # else:
    #     d = copy.deepcopy(d)
    #     d["epsilon"] = epsmin
    #     d["epsilon"] = d["epsilon"]
    # prob["design_config"]["void"] = d

    prob["iters"] = iters
    save_problem(prob, path)
    return prob


def Target(
    key,
    target,
    weight,
    param="T",
    frequency=None,
    wl_1_f=None,
    wavelength=None,
    func="abs",
):
    if frequency is not None:
        wavelength = wl_1_f / frequency
    return {
        "param": param,
        "key": str(key),
        "wavelength": wavelength,
        "target": target,
        "weight": weight,
        "func": func,
    }


def apply_design(c0, sol):
    path = sol["path"]
    a = gf.Component()
    a.add_ref(c0)
    fill = sol["design_config"]["fill"]["layer"]
    dl = sol["design_config"]["DESIGN"]
    for i, d in enumerate(sol["designs"]):
        x0, y0 = d["bbox"][0]
        x1, y1 = d["bbox"][1]
        b = gf.Component()
        b.add_polygon([(x0, y0), (x1, y0), (x1, y1), (x0, y1)], layer=fill)
        a = gf.boolean(a, b, "not", layer=fill)
    c = gf.Component()
    c << a
    # for layer in c0.layers:
    #     if layer != dl:
    #         c.add_ref(c0.extract([layer]))
    # c.show()
    # raise ValueError()
    g = gf.import_gds(os.path.join(path, f"optimized_design_{i+1}.gds"))
    polygons = g.get_polygons(merge=True)
    g = gf.Component()
    for p in polygons[1]:
        g.add_polygon(p, layer=fill)
    g = c << g
    g.xmin = x0
    g.ymin = y0
    return c



import gdsfactory as gf
from gdsfactory.generic_tech.layer_map import LAYER
from gdsfactory.technology import LogicalLayer, LayerStack, LayerLevel
import gdsfactory as gf
from gdsfactory.generic_tech import LAYER, LAYER_STACK
from gdsfactory.generic_tech.get_klayout_pyxs import get_klayout_pyxs
from gdsfactory.technology import LayerLevel, LayerStack, LayerViews, LayerMap
from gdsfactory.generic_tech import get_generic_pdk

from gdsfactory.config import PATH
from gdsfactory.technology import (
    LayerLevel,
    LayerStack,
    LayerView,
    LayerViews,
    LayerMap,
)

Layer = tuple[int, int]

# gf.generic_tech.


class MyLayerMap(LayerMap):
    WG: Layer = (1, 0)
    WGCLAD: Layer = (111, 0)
    BOX: Layer = (112, 0)
    PORT: Layer = (1, 10)
    DESIGN: Layer = (200, 0)
    GUESS: Layer = (201, 0)


LAYER = MyLayerMap

nm = 1e-3
thickness_wg = 220 * nm
thickness_slab_deep_etch = 90 * nm
thickness_slab_shallow_etch = 150 * nm
box_thickness = 0.5
thickness_clad = 0.5

layer_core = LogicalLayer(layer=LAYER.WG)
layer_clad = LogicalLayer(layer=LAYER.WGCLAD)
layer_box = LogicalLayer(layer=LAYER.BOX)
DESIGN = LogicalLayer(layer=DESIGN)

LAYER_STACK.layers.update(
    dict(
        core=LayerLevel(
            layer=layer_core,
            thickness=thickness_wg,
            zmin=0,
            material="Si",
            mesh_order=2,
        ),
        box=LayerLevel(
            layer=layer_box,
            thickness=box_thickness,
            zmin=-box_thickness,
            material="SiO2",
            mesh_order=9,
        ),
        clad=LayerLevel(
            layer=layer_clad,
            zmin=0.0,
            material="SiO2",
            thickness=thickness_clad,
            mesh_order=10,
        ),
        design=LayerLevel(
            layer=DESIGN,
            zmin=0.0,
            material=None,
            thickness=0.0001,
            mesh_order=100,
        ),
    )
)

pdk = get_generic_pdk()
pdk.activate()
LAYER_VIEWS = pdk.layer_views



from scipy.special import comb
from ..constants import *
from ..utils import *
import gdsfactory as gf
from gdsfactory.cross_section import Section
from networkx import center
from numpy import cumsum
from gdsfactory.generic_tech import LAYER_STACK, LAYER

# import .utils as utils
# from layers import *


def port_bbox(p):
    c = np.array(p.center)
    v = (
        p.width
        / 2
        * np.array(
            [
                np.cos(np.radians(p.orientation + 90)),
                np.sin(np.radians(p.orientation + 90)),
            ]
        )
    )
    return [(c - v).tolist(), (c + v).tolist()]


def bezier_curve(points, num_points=100):
    n = len(points) - 1
    t = np.linspace(0, 1, num_points)
    curve_points = np.zeros((num_points, 2))

    for i, t_val in enumerate(t):
        for j, point in enumerate(points):
            curve_points[i] += (
                comb(n, j) * (t_val**j) * ((1 - t_val) ** (n - j)) * point
            )

    return curve_points


def mimo(
    l,
    w,
    west=0,
    east=0,
    south=0,
    north=0,
    width_wg=0.5,
    length_wg=None,
    wwg_west=None,
    wwg_east=None,
    wwg_south=None,
    wwg_north=None,
    taper=0,
    init_geometry=None,
    layer=WG,
    design_layer=DESIGN,
    centered=False,
    **kwargs,
):
    if not length_wg:
        length_wg = lwg_out = 2 * width_wg
    design = gf.Component()
    c = gf.Component(**kwargs)

    if centered:
        xmin = -l / 2
        ymin = -w / 2
        xmax = l / 2
        ymax = w / 2
        p = [(-l / 2, -w / 2), (l / 2, -w / 2), (l / 2, w / 2), (-l / 2, w / 2)]
    else:
        xmin = 0
        ymin = 0
        xmax = l
        ymax = w
        p = [(0, 0), (l, 0), (l, w), (0, w)]
    design.add_polygon(p, layer=design_layer)
    # c.add_polygon(p,                       layer=layer)

    port_pos_sides = [west, north, east, south]
    for i, v, d in zip(range(4), port_pos_sides, [w, l, w, l]):
        if type(v) is int:
            port_pos_sides[i] = [(0.5 + j) * d / v for j in range(v)]

    wwg_sides = [wwg_west, wwg_east, wwg_south, wwg_north]
    for i, v in enumerate(wwg_sides):
        n = len(port_pos_sides[i])
        if v is None:
            v = width_wg
        if type(v) is float or type(v) is int:
            wwg_sides[i] = [v] * n

    nports = 0
    for i, x, y, ds, wwgs, orientation in zip(
        range(4),
        [xmin, xmin, xmax, xmax],
        [ymin, ymax, ymax, ymin],
        port_pos_sides,
        wwg_sides,
        [180, 90, 0, -90],
    ):
        for width_wg, d in zip(wwgs, ds):
            center = [
                x + cos(np.radians(orientation - 90)) * d,
                y + sin(np.radians(orientation - 90)) * d,
            ]
            nports += 1
            name = "o" + str(nports)
            wwg2 = (1 + taper) * width_wg
            design.add_port(
                name, center=center, width=wwg2, orientation=orientation, layer=layer
            )
            wg = c << gf.components.taper(
                length=length_wg, width1=width_wg, width2=wwg2, layer=layer
            )
            wg.connect("o2", design.ports[name])
            c.add_port(name, port=wg.ports["o1"])
    design = c << design
    # p = []
    # for i in range(nports):
    #     # for i in [0]:
    #     pi = design.ports[f'o{i+1}']
    #     pj = design.ports[f'o{((i+1) % nports)+1}']
    #     a, _ = np.array(port_bbox(pi))
    #     _, b = np.array(port_bbox(pj))
    #     n1 = - np.array([cos(np.radians(pi.orientation)),
    #                      sin(np.radians(pi.orientation))])
    #     n2 = -np.array([cos(np.radians(pj.orientation)),
    #                    sin(np.radians(pj.orientation))])
    #     v = b-a
    #     d = np.linalg.norm(v)
    #     n = v/d

    #     # l=[a, a+.3*d * (.5*n+.5*n1),  b+.3*d*(-.5*n+.5*n2), b]
    #     l = [a, a+.5*d * n1,  b+.5*d*n2, b]
    #     p.extend(bezier_curve(l))

    # p.extend([a+s*d * (2*s*n+(1-2*s)*n1) for s in np.linspace(0, .5, 50)])
    # p.extend(reversed([b+s*d * (-2*s*n+(1-2*s)*n2)
    #          for s in np.linspace(0, .5, 50)]))
    if type(init_geometry) is list:
        for i, j in init_geometry:
            pi = design.ports[f"o{i}"]
            pj = design.ports[f"o{j}"]
            a = np.array(pi.center)
            b = np.array(pj.center)
            n1 = -np.array(
                [cos(np.radians(pi.orientation)), sin(np.radians(pi.orientation))]
            )
            n2 = -np.array(
                [cos(np.radians(pj.orientation)), sin(np.radians(pj.orientation))]
            )
            v = b - a
            d = np.linalg.norm(v)
            n = v / d
            l = [a, a + 0.5 * d * n1, b + 0.5 * d * n2, b]

            c << gf.components.bends.bezier(
                [x.tolist() for x in l], allow_min_radius_violation=True, width=pi.width
            )
    if init_geometry == 1:
        c.add_polygon(p, layer=layer)
    p = [
        {
            "name": v.name,
            "center": v.center,
            "width": v.width,
            "orientation": v.orientation,
        }
        for v in design.ports
    ]
    c.info["ports"] = p
    return c



from copy import deepcopy
from gdsfactory.generic_tech import get_generic_pdk
from gdsfactory.technology import (
    LayerLevel,
    LayerStack,
    LayerView,
    LayerViews,
    LayerMap,
)

pdk = get_generic_pdk()
pdk.activate()
LAYER_VIEWS = pdk.layer_views

eps0 = 8.854187817e-12
Z0 = 377 * (3e8) * eps0



from .setup import *
from ..constants import *
from ..layers import *
from ..utils import *
import gdsfactory as gf
from copy import deepcopy
import matplotlib.pyplot as plt
import numpy as np
import math

from gdsfactory.cross_section import Section
from gdsfactory.generic_tech import LAYER_STACK, LAYER


def make_prob(
    path,
    component=None,
    wavelengths=None,
    characteristic_wavelength=None,
    entries=None,
    keys=["2,1"],
    study="sparams",
    nres=4,
    sources=None,
    monitors=None,
    frequencies=None,
    wl_1_f=None,
    modes=None,
    #
    layer_stack={},
    materials=MATERIALS,
    verbose=True,
    #
    targets=None,
    symmetries=[],
    # design_layer=DESIGN,
    init_pattern_level=None,
    init_pattern_spacing=None,
    init_pattern_diameter=None,
    area_change=0.01,
    fill_material=None,
    void_material=None,
    iters=None,
    lmin=None,
    stoploss=None,
    design_ports=None,
    #
    path_length_multiple=0.85,
    approx_2D_mode=None,
    pixel_size=0.01,
    info=None,
    **kwargs
):
    if approx_2D_mode:
        N = 2
    else:
        N = 3

    if wl_1_f is None:
        ordering = "wavelength"
    else:
        ordering = "frequency"

    if targets is not None:
        study = "inverse_design"
        design_layer = layer_stack["design"].layer.layer
        if "relative_mesh_density" not in materials["design"]:
            materials["design"]["relative_mesh_density"] = math.sqrt(
                materials[fill_material]["epsilon"]
            )

        keys = SortedSet()
        b = wavelengths is None
        if b:
            wavelengths = SortedSet()
        for d in targets:
            if b:
                wavelengths.add(d["wavelength"])
            keys.add(d["key"])
        keys = list(keys)

    if frequencies is not None:
        assert wl_1_f is not None
        assert wavelengths is None
        wavelengths = [wl_1_f / f for f in frequencies]

    if type(wavelengths) in [int, float]:
        wavelengths = [wavelengths]
    wavelengths = sorted(wavelengths)
    if characteristic_wavelength is None:
        characteristic_wavelength = median(wavelengths)
    # nres *= characteristic_wavelength / min(map(min, wavelengths))

    if not entries:
        entries = []

        for w in wavelengths:
            for k in keys:
                try:
                    entries.append([w, *unpack_sparam_key(k)])
                except Exception as e:
                    0
    l = []
    for w, po, mo, pi, mi in entries:
        k = [w, pi, mi, pi, mi]
        if k not in entries:
            l.append(k)
    entries.extend(l)

    imow = {}
    for w, po, mo, pi, mi in entries:
        if pi not in imow:
            imow[pi] = {}
        if mi not in imow[pi]:
            imow[pi][mi] = {}

        if po not in imow[pi][mi]:
            imow[pi][mi][po] = mo
        else:
            imow[pi][mi][po] = max(imow[pi][mi][po], mo)

    runs = []
    for _w in [1]:
        for i in imow:
            for mi in imow[i]:
                d = {
                    "sources": {
                        i: {
                            "wavelength_mode_numbers": [[w, [mi]] for w in wavelengths],
                        }
                    },
                    "monitors": {
                        o: {
                            "wavelength_mode_numbers": [
                                [w, list(range(imow[i][mi][o] + 1))]
                                for w in wavelengths
                            ],
                        }
                        for o in imow[i][mi]
                    },
                }
                d["sources"] = SortedDict(d["sources"])
                d["monitors"] = SortedDict(d["monitors"])
                if component is None:
                    0
                else:
                    runs.append(d)

    prob = setup(
        path,
        component=component,
        study=study,
        characteristic_wavelength=characteristic_wavelength,
        wl_1_f=wl_1_f,
        wavelengths=wavelengths,
        runs=runs,
        layer_stack=layer_stack,
        materials=materials,
        modes=modes,
        verbose=verbose,
        keys=keys,
        ordering=ordering,
        nres=nres,
        approx_2D_mode=approx_2D_mode,
        pixel_size=pixel_size,
        #  sources=sources, monitors=monitors,
        **kwargs
    )

    if targets:
        polys = component.extract([design_layer]).get_polygons()
        if not symmetries:
            symmetries = [[]] * len(polys)
        else:
            if type(symmetries[0]) is not list:
                symmetries = [symmetries] * len(polys)

        def _bbox(b):
            return [[b.left / 1e3, b.bottom / 1e3], [b.right / 1e3, b.top / 1e3]]

        ks = set(materials[fill_material].keys()).intersection(
            set(materials[void_material].keys())
        )
        swaps = {
            k: (materials[void_material][k], materials[fill_material][k]) for k in ks
        }
        designs = []
        for p, s in zip(list(polys.values())[0], symmetries):
            bbox = _bbox(p.bbox())
            if N == 3:
                l = get_layers(layer_stack, design_layer)[0]
                bbox[0].append(l.zmin)
                bbox[1].append(l.zmin + l.thickness)
            designs.append(
                {
                    "bbox": bbox,
                    "swaps": swaps,
                    "symmetries": s,
                }
            )
        prob = {
            **prob,
            **{
                "targets": targets,
                "designs": designs,
                "init_pattern_level": init_pattern_level,
                "init_pattern_spacing": init_pattern_spacing,
                "init_pattern_diameter": init_pattern_diameter,
                "area_change": area_change,
                "iters": iters,
                "lmin": lmin,
                "stoploss": stoploss,
                "path_length_multiple": path_length_multiple,
                "design_ports": design_ports,
            },
        }
    save_problem(prob, path)
    return prob

    # l = [k for k in imow if port_number(k) == pi]
    # if not l:
    #     imow[f"o{pi}@{mi}"] = []
    # else:
    #     k = l[0]
    #     mn = max(mode_number(k), mi)
    #     if mn != mode_number(k):
    #         imow[i] = imow[k]
    #         del imow[k]

    # l = [k for k in imow[i] if port_number(k) == po]
    # if not l:
    #     imow[f"o{pi}@{mi}"]
    # else:
    #     k = l[0]
    #     mn = max(mode_number(k), mi)
    #     if mn != mode_number(k):
    #         imow[f"o{pi}@{mn}"] = imow[k]
    #         del imow[k]

    # if po not in imow[pi]:
    #     imow[pi]["o"][po] = mo
    # else:
    #     imow[pi]["o"][po] = max(imow[pi]["o"][po], mo)
